"""
File writing stuff
"""
import html
from typing import Any, Dict, List

import html2text

from so_pip.make_from_template import load_template

WINDOWS_LINE_ENDING = "\r\n"
UNIX_LINE_ENDING = "\n"


def format_comments(comments: List[Dict[str, Any]]) -> List[str]:
    """Put comments into ordinary strings"""
    comments_formatted = []
    for comment in comments:
        owner = comment["owner"].get("display_name", "(no user)")
        license_text = comment["content_license"]
        comments_formatted.append(owner + "," + license_text)
    return comments_formatted


def write_as_html(post: Dict[str, Any], submodule_name: str) -> None:
    """Dump post in readable form."""
    template = load_template("post.html.jinja", autoescape=False)
    data = {
        "title": post.get("title", ""),
        "content": post["body"],
        "comments": format_comments(post.get("comments", [])),
    }
    result = template.render(item=data)
    with open(submodule_name + ".html", "w", encoding="utf-8") as diagnostics:
        diagnostics.write(result)


def write_as_text(post: Dict[str, Any], submodule_name: str) -> None:
    """Dump post in readable form."""
    with open(submodule_name + ".txt", "w", encoding="utf-8") as diagnostics:
        text = html2text.html2text(post["body"])
        diagnostics.write(text)


def write_as_md(post: Dict[str, Any], submodule_name: str) -> None:
    """Dump post in readable form."""
    if "body_markdown" not in post:
        return

    markdown = post["body_markdown"]

    with open(submodule_name + ".md", "w", encoding="utf-8") as diagnostics:
        # https://stackoverflow.com/a/43678795/33264
        markdown = markdown.replace(WINDOWS_LINE_ENDING, UNIX_LINE_ENDING)
        # don't know why it has html escapes, possibly a security feature
        # for when you know you're going to put it into HTML?
        markdown = html.unescape(markdown)
        diagnostics.write(markdown)



"""
Scrape twitter & github urls

https://stackapps.com/questions/7549/
how-can-i-get-user-links-in-the-api-call-like-twitter-github-and-so-on
"""
from functools import lru_cache
from typing import Tuple

import requests
from bs4 import BeautifulSoup


@lru_cache(maxsize=1000)
def scrape_urls(user_id: int) -> Tuple[str, str]:
    """Get twiter and github"""
    response = requests.get(f"https://stackoverflow.com/users/{user_id}/")
    if response.status_code != 200:
        return "", ""
    soup = BeautifulSoup(response.text, features="html.parser")
    twitter = ""
    github = ""
    for anchor in soup.findAll("a"):
        if "href" in anchor.attrs:
            candidate = anchor.attrs["href"]
            if (
                "twitter.com" in candidate
                and candidate != "https://twitter.com/stackoverflow"
            ):
                twitter = candidate
            if "github.com" in candidate:
                github = candidate
        # other URL we can get from API
        if twitter and github:
            return twitter, github
    return twitter, github


# print(scrape_urls(2145778))



"""
Settings, eventually will be docopts
"""
import ast

# do we want this question/post at all?
import configparser
import os
from typing import Tuple, cast

from so_pip.utils.files_utils import find_file

if os.path.exists(".so_pip.ini"):
    # exist where user executes from.
    CONFIG_PATH = ".so_pip.ini"
else:
    # unit tests, etc.
    CONFIG_PATH = find_file("../.so_pip.ini", __file__)
if os.path.exists(CONFIG_PATH):
    print(f"Loading config from {CONFIG_PATH}")
    config = configparser.ConfigParser()
    result = config.read(CONFIG_PATH)
else:
    print("Did not fine a .so_pip.ini file anywhere, using default behaviors.")
    # duck typing
    config = cast(configparser.ConfigParser, configparser.RawConfigParser())
    config.add_section("POST_FILTERS")
    config.add_section("VENDORIZING")
    config.add_section("CODE_CLEANUP")
    config.add_section("SUPPORTING_FILES")
    config.add_section("LANGUAGE_DETECTION")
    config.add_section("SHELL_CONFIG")

section = config["POST_FILTERS"]
MINIMUM_SCORE = ast.literal_eval(section.get("MINIMUM_SCORE", "0"))
KEEP_ANSWERS_WITH_NO_CODE = ast.literal_eval(
    section.get("KEEP_ANSWERS_WITH_NO_CODE", "False")
)
KEEP_ANSWERS_WITH_NO_DEF_OR_CLASS = ast.literal_eval(
    section.get("KEEP_ANSWERS_WITH_NO_CODE", "False")
)
KEEP_ANSWERS_WITH_THESE_LANGUAGES = ast.literal_eval(
    section.get("KEEP_ANSWERS_WITH_THESE_LANGUAGES", '["*"]')
)

# this pattern won't work once so_pip is pip installed.
section = config["VENDORIZING"]

# {target_folder}/{package_prefix}_{package_name}/{package_info}.py
# so_pip_packages/fizzbuzz_question_forest_cake/

# how to make the module look
# ------------------------
section = config["CODE_CLEANUP"]
COMMENT_OUT_BAD_PYTHON = ast.literal_eval(
    section.get("COMMENT_OUT_BAD_PYTHON", "False")
)
# one liners are often english with <code/> used for styling
ASSUME_ONE_LINER_IS_NOT_CODE = ast.literal_eval(
    section.get("ASSUME_ONE_LINER_IS_NOT_CODE", "True")
)
WRAP_IN_RUN = ast.literal_eval(section.get("WRAP_IN_RUN", "False"))
TEXT_AS_COMMENTS = ast.literal_eval(section.get("TEXT_AS_COMMENTS", "True"))
# Slow but necessary because so many answers are old.
BUMP_TO_PY3 = ast.literal_eval(section.get("BUMP_TO_PY3", "True"))
IMPORT_STARTS_NEW_FILE = ast.literal_eval(section.get("IMPORT_STARTS_NEW_FILE", "True"))

section = config["SUPPORTING_FILES"]
POSTS_AS_TXT = ast.literal_eval(section.get("POSTS_AS_TXT", "True"))
POSTS_AS_HTML = ast.literal_eval(section.get("POSTS_AS_HTML", "True"))
POSTS_AS_MD = ast.literal_eval(section.get("POSTS_AS_MD", "True"))
METADATA_IN_INIT = ast.literal_eval(section.get("METADATA_IN_INIT", "True"))
GENERATE_REQUIREMENTS_TXT = ast.literal_eval(
    section.get("GENERATE_REQUIREMENTS_TXT", "True")
)
GENERATE_AUTHORS = ast.literal_eval(section.get("GENERATE_AUTHORS", "True"))
GENERATE_README = ast.literal_eval(section.get("GENERATE_README", "True"))
GENERATE_CODE_OF_CONDUCT = ast.literal_eval(
    section.get("GENERATE_CODE_OF_CONDUCT", "True")
)
GENERATE_CHANGE_LOG = ast.literal_eval(section.get("GENERATE_CHANGE_LOG", "True"))
GENERATE_SETUP_CFG = ast.literal_eval(section.get("GENERATE_SETUP_CFG", "True"))

section = config["LANGUAGE_DETECTION"]
DEFAULT_LANGUAGE = cast(
    Tuple[str, str],
    ast.literal_eval(section.get("DEFAULT_LANGUAGE", "('.py', 'python')")),
)

# Language guesser needs hints.
# Default hints for python and javascript because
# those languages now (or will) have more support for extracting a package.
POSSIBLE_LANGUAGES = ast.literal_eval(section.get("POSSIBLE_LANGUAGES",
                                                  "['python','javascript']"))

# cli client stuff
section = config["SHELL_CONFIG"]
RUNNING_IN_VENV = "VIRTUAL_ENV" in os.environ
SHELL = section.get("SHELL", "") if not RUNNING_IN_VENV else ""
QUIET = False



"""
Abstract model of the package_info I'm extracting from an post.
"""
from dataclasses import dataclass, field
from typing import List

from so_pip.api_clients.what_that_code_facade import guess_language_and_extension

from so_pip.parse_python.python_validator import validate_python


@dataclass()
class CodeBlock:
    """Represents an html code block"""

    # raw text, straight from the MD/HTML
    raw_text: str = ""
    # Cleaned up source code
    code_text: str = ""
    # English code that will need to be commented out
    header_comments: str = ""
    footer_comments: str = ""
    # If not, could still be bad python or English or some other language
    is_valid_python: bool = False
    errors: List[str] = field(default_factory=list)
    is_python_error_message: bool = False
    is_interactive_block: bool = False
    is_ipython_block: bool = False
    starts_new_file: bool = False
    extension: str = ""
    language: str = ""

    def analyze(self, tags: List[str]) -> None:
        """Do expensive checks"""
        self.is_ipython_block = ">>>" in self.code_text
        self.is_valid_python, self.errors = validate_python(self.code_text)

        value = guess_language_and_extension(self.code_text,
                                             surrounding_text=self.header_comments + self.footer_comments,
                                             tags=tags)
        self.extension, self.language = value



"""
Utility functions to help with code coverage.

Imagine:

def foo(bar):
    if bar is None:
        raise Exception()

Hitting the guard requires an extra unit test and provides little
value to get 100% coverage. (And to hit 90% coverage, you need to
have 100% coverage on most files)
"""
import logging
from typing import Any, Dict

LOGGER = logging.getLogger(__name__)


def must_not_be_none(value: Any, message: str = "Value must not be none") -> None:
    """
    Raise exception if value is None
    """
    if value is None:
        LOGGER.error(f"Can't be none, but got {value}")
        raise TypeError(message)


def must_not_be_truthy(value: bool, message: str = "Value must not be truthy") -> None:
    """
    Guard to assert truthiness & make it easier to get unit test coverage.
    """
    if value:
        LOGGER.error(f"Can't be truthy, but got {value}")
        raise TypeError(message)


def must_be_truthy(value: Any, message: str = "Value must not be truthy") -> None:
    """
    Guard to assert truthiness & make it easier to get unit test coverage.
    """
    if not value:
        LOGGER.error(f"Must be truthy, but got {value}")
        raise TypeError(message)


def must_not_be_falsy(value: Any, message: str = "Value must not be falsy") -> None:
    """
    Raise exception if value is falsy (i.e. False, "", None, [], etc)
    """
    if not value:
        LOGGER.error(f"Must not be falsy, but got {value}")
        raise TypeError(message)


def must_be_falsy(value: Any, message: str = "Value must not be falsy") -> None:
    """
    Raise exception if value is not falsy (i.e. False, "", None, [], etc)
    """
    if value:
        LOGGER.error(f"Must be falsy, but got {value}")
        raise TypeError(message)


def must_not_already_be_in_dictionary(
    key: Any,
    dictionary: Dict[Any, Any],
    message: str = "Key already in dictionary",
    warn_only: bool = False,
) -> None:
    """Assert we aren't about to insert a key twice to a dict"""
    if key in dictionary:
        if warn_only:
            print(
                f"{key} is already in dictionary where it has "
                f"a value of {dictionary[key]}"
            )
        else:
            raise TypeError(message)


def assert_is_number(value: str) -> None:
    """
    Raise error if there is a value and the value is not a number
    """
    if value is None:
        # Why did I do this?
        return
    # pylint: disable=broad-except
    # noinspection PyBroadException
    try:
        _ = int(value)
    except BaseException:
        LOGGER.error(f"Expected numerical id, got {value}")



"""
Is this valid python, what version?

Valid syntax is valid.
Invalid syntax could still be python
Invalid syntax could also be something that isn't python at all.
"""
import ast
import subprocess  # nosec
from typing import List, Tuple

from so_pip.cli_clients.external_commands import pyflakes, vermin


def validate_with_vermin(folder: str) -> str:
    """Infer python versions"""
    result = vermin(folder)
    # "Minimum required versions: 2.0  Incompatible versions:     3"
    # ~2       No known reason it won't work with py2.
    # !2       It is known that it won't work with py2.
    # 2.5, !3  Works with 2.5+ but it is known it won't work with py3.
    # ~2, 3.4  No known reason it won't work with py2, works with 3.4+"
    # print(result)
    if "~2, ~3" in result:
        return "*"
    if "2.0, 3.0" in result:
        return "*"
    # pylint: disable=broad-except
    # noinspection PyBroadException
    try:
        parts = result.split(":")
        minimum = parts[1].strip().split(" ")[0].strip("\n")
        if "\n" in minimum:
            minimum = minimum.split("\n")[0]
        if "," in minimum:
            minimum = minimum.split(",")
        if "Incompatible" in result:
            parts = result.split("Incompatible versions:")
            maximum = parts[1].strip("\n").strip(" ")
            maximum = f"!{maximum}"
        else:
            maximum = ""
    except BaseException:
        print("unexpected compat string")
        print(result)
        raise
    if maximum:
        return f">={minimum}, {maximum}"
    return f">={minimum}"


def validate_python(code: str) -> Tuple[bool, List[str]]:
    """Validate by ast parsing"""
    try:
        _ = ast.parse(code)
        # same, but also writes files, I think
        # __ = compile(code, 'blah.py', 'exec', dont_inherit=True)
        return True, []
    except SyntaxError as syntax_error:
        error_message = f"{str(syntax_error)}"
    return False, [error_message]


def validate_with_pyflakes(file_name: str) -> Tuple[bool, List[str]]:
    """Does pyflakes think this is valid python?"""
    # pyflakes doesn't recognize latest versions of py3.x
    try:
        results = pyflakes(file_name)
        if "syntax" in results:
            return False, results.split("/")
    except subprocess.CalledProcessError as cpe:
        return False, [str(cpe)]
    return True, []



"""
Abstract model of the package_info I'm extracting from an post.
"""
from dataclasses import dataclass, field
from typing import List

from so_pip.api_clients.what_that_code_facade import guess_language_and_extension
from so_pip.models.code_block_model import CodeBlock
from so_pip.parse_code.comment_out_anything import html_to_comments
from so_pip.parse_python.python_validator import validate_python


@dataclass()
class CodeFile:
    """Represents an file which has many code blocks and comments"""

    doc_string: str = ""
    file_name: str = ""
    extension: str = ""
    language: str = ""

    is_valid_python: bool = False
    errors: List[str] = field(default_factory=list)

    code_blocks: List[CodeBlock] = field(default_factory=list)

    failed_parse: bool = False

    def to_write(self) -> List[str]:
        """Lines to write, including comments"""
        if not self.code_blocks:
            raise TypeError("No blocks")
        items = []



        for block in self.code_blocks:
            if block.header_comments:
                items.append(html_to_comments(block.header_comments, block.language))
            if block.code_text:
                items.append(block.code_text)
            if block.footer_comments:
                items.append(html_to_comments(block.footer_comments, block.language))
        if self.code_blocks and not items:
            raise TypeError("We have code blocks, but no lines of code?")
        return items

    def all_code(self) -> List[str]:
        """Just code"""
        return [block.code_text for block in self.code_blocks]

    def all_non_code(self) -> List[str]:
        """Just non code"""
        return [block.header_comments for block in self.code_blocks] + \
               [block.footer_comments for block in self.code_blocks]

    def non_comment_lines(self) -> int:
        """Count code lines"""
        return sum(1 for x in self.all_code() if len(x) and not x.startswith("#"))

    def preview_final_code(self) -> str:
        """For checking if a word is in the code for an post"""
        return "\n".join(self.all_code())

    def preview_non_code(self) -> str:
        """Looking for clues about what the code language is"""
        return "\n".join(self.all_non_code())

    def analyze(self, tags: List[str]) -> None:
        """Do expensive checks"""
        final_code = self.preview_final_code()
        final_context = self.preview_non_code()

        self.is_valid_python, self.errors = validate_python(final_code)

        pair = guess_language_and_extension(final_code,
                                            surrounding_text=final_context,
                                            tags=tags)
        self.extension, self.language = pair
        if not self.extension:
            self.extension = ".wtf"



"""
SO questions and answers look a lot alike jupyter notebooks.
"""
import json

import nbformat


def proof_of_concept() -> None:
    """This is not ready for prime time."""
    thing = nbformat.v4.new_notebook()

    cell = nbformat.v4.new_code_cell(source="print('hello')")
    md_cell = nbformat.v4.new_markdown_cell(
        source="""Things are looking up
    --------
    _are_ *they* not?
    """,
    )
    thing.metadata = {
        "kernelspec": {
            "display_name": "Python 3",
            "language": "python",
            "name": "python3",
        },
        "language_info": {
            "codemirror_mode": {"name": "ipython", "version": 2},
            "file_extension": ".py",
            "mimetype": "text/x-python",
            "name": "python",
            "nbconvert_exporter": "python",
            "pygments_lexer": "ipython2",
            "version": "2.7.6",
        },
    }

    thing.cells = [cell, md_cell]
    with open("thing.ipynb", "w", encoding="utf-8") as file:
        string = json.dumps(thing)
        file.write(string)



"""
Generate a CHANGELOG type file based on the edit log in SO

Design strategy, follow this format: https://keepachangelog.com/en/1.0.0/

Focus on changes to the question/answers *only*, not who did it.
"""
from datetime import datetime
from typing import Any, Dict, List

import markdown

from so_pip.api_clients.stackapi_facade import get_json_revisions_by_post_id
from so_pip.make_from_template import load_template
from so_pip.models.authors_model import normalize_user_link


def changelog_for_post(post: Dict[str, Any], package_folder: str) -> None:
    """Requirements for running `safety`"""
    versions = []
    post_id = post["answer_id"] if "answer_id" in post else post["question_id"]
    revision_json = get_json_revisions_by_post_id(post_id)
    # if len(revision_json.get("items", [])) == 0:
    #     return
    for revision in revision_json.get("items", []):
        # if "content_license" not in revision:
        #     print("but why")
        if "revision_number" not in revision:
            # happens with protected posts
            # no revision number, no version number
            continue
        utc_date = datetime.utcfromtimestamp(revision["creation_date"]).isoformat()

        revision_number = revision["revision_number"]
        what = "Added" if revision_number == 1 else "Changed"
        version = {
            "version": f"0.1.{revision_number}",
            "date": f"{utc_date}",
            "changes": {
                "content_license": revision.get("content_license", ""),
                "user": "",
                "comment": "Created post",
                "what": what,
            },
        }
        if revision["user"]["user_type"] != "does_not_exist":
            display_name = revision["user"]["display_name"]
            url = revision["user"]["link"]
            link = normalize_user_link(url, revision["user"]["user_id"])
            version["changes"]["user"] = f"{display_name} <{link}>"
        else:
            version["changes"]["user"] = "No user."
        if revision_number == 1:
            added = "Added answer" if "answer_id" in post else "Added Question"
            version["changes"]["comment"] = added
        else:
            version["changes"]["comment"] = revision.get("comment", "")
        version[
            "revision_url"
        ] = f"https://stackoverflow.com/revisions/{post_id}/{revision_number}"
        versions.append(version)

    with open(
        package_folder + "/CHANGELOG.md", "w", encoding="utf-8", errors="replace"
    ) as log:
        text = render_change_log(data=versions)

        # calling this to see if it is somewhat valid markdown
        _ = markdown.markdown(text)
        log.write(text)


def render_change_log(
    data: List[Dict[str, Any]],
) -> str:
    """
    Databind to jinja file
    """
    template = load_template(template_filename="HISTORY.md.jinja", autoescape=False)
    output_text = template.render(item=data)
    return output_text



"""
Generate a list of 10,000 good words.
"""


def run() -> None:
    """Combine lots of words, remove lots of bad words"""
    # Typo: In word 'acer'
    with open("bad_words.txt") as bad:
        bad_words = [
            _.strip(" ,*.") for _ in bad.read().split("\n") if not _.startswith("#")
        ]

    with open("wordlist.10000.txt") as long_file:
        ten_k = [
            _.strip(" ,*.")
            for _ in long_file.read().split("\n")
            if not _.startswith("#")
        ]
        for word in ten_k:
            if len(word) == 1:
                # letters
                bad_words.append(word)
            if not word.isalpha():
                # letters
                bad_words.append(word)
            if word.upper() == word:
                # accronyms
                bad_words.append(word)
            if len([letter for letter in word if letter in "aeiou"]) == 0:
                bad_words.append(word)
    with open("three_k.txt") as three_file:
        three_k = [
            _.strip(" ,*.")
            for _ in three_file.read().split("\n")
            if not _.startswith("#")
        ]
        for word in three_k:
            if len(word) == 1:
                # letters
                bad_words.append(word)
            if not word.isalpha():
                # letters
                bad_words.append(word)
            if word.upper() == word:
                # accronyms
                bad_words.append(word)
            if len([letter for letter in word if letter in "aeiou"]) == 0:
                bad_words.append(word)

    with open("most-common-nouns-english.csv") as most_common_file:
        most_common = [
            _.strip(" ,*.")
            for _ in most_common_file.read().split("\n")
            if not _.startswith("#")
        ]
        for word in most_common:
            if len(word) == 1:
                # letters
                bad_words.append(word)
            if not word.isalpha():
                # letters
                bad_words.append(word)
            if word.upper() == word:
                # accronyms
                bad_words.append(word)
            if len([letter for letter in word if letter in "aeiou"]) == 0:
                bad_words.append(word)

    most_common_simpler = []
    for word in most_common:
        if "," in word:
            parts = word.split(",")
            first = parts[0]
            most_common_simpler.append(first)
        else:
            most_common_simpler.append(word)

    with open("people_names.txt") as people_file:
        people = [
            _.strip(" ,*.")
            for _ in people_file.read().split("\n")
            if not _.startswith("#")
        ]
        people = people[:1000]
    dedupe = {
        word.strip(", .").lower()
        for word in set(ten_k + three_k + most_common_simpler + people)
    }
    new_set = [word for word in dedupe if word not in bad_words]
    print(len(new_set))
    list.sort(new_set)
    print(new_set)
    with open("clean_ten_k.txt", "w", encoding="utf-8") as clean:
        clean.write("\n".join(list(new_set[:10000])))


if __name__ == "__main__":
    run()



"""
Sett up the whats_that_code call
"""
from typing import List, Tuple

from so_pip.settings import POSSIBLE_LANGUAGES
from whats_that_code.election import guess_language_all_methods
from whats_that_code.known_languages import FILE_EXTENSIONS


def guess_language_and_extension(code: str,
                                 surrounding_text: str = "",
                                 tags: List[str] = None) -> Tuple[str, str]:
    """Guess language and extension"""
    if not code:
        return "", ""

    result = guess_language_all_methods(code, file_name="",
                                        # TODO: include this when bug fixed.
                                        #surrounding_text=surrounding_text,
                                        surrounding_text="",
                                        tags=tags,
                                        priors=POSSIBLE_LANGUAGES)
    if result:
        if result in FILE_EXTENSIONS:
            # take top extension option
            return FILE_EXTENSIONS[result][0], result
        return f".{result}", result

    return "", ""



"""
Low level access. Makes no effort to parse results.
"""

import os
from functools import lru_cache
from typing import Any, Dict, Tuple, cast

from stackapi import StackAPI

if os.environ.get("key", None):
    SITE = StackAPI("stackoverflow", key=os.environ["key"])
else:
    SITE = StackAPI("stackoverflow")


@lru_cache(maxsize=1000)
def get_json_by_search(query: str, tagged: Tuple[str, ...]) -> Dict[str, Any]:
    """Low level access, returns unprocessed json"""
    return cast(
        Dict[str, Any],
        SITE.fetch(
            "search?tagged={tagged}&intitle={intitle}",
            tagged=[";".join(tagged)],
            intitle=[query],
        ),
    )


@lru_cache(maxsize=1000)
def get_json_by_question_id(question_id: int) -> Dict[str, Any]:
    """Low level access, returns unprocessed json"""
    return cast(
        Dict[str, Any],
        SITE.fetch(
            "questions/{ids}",
            ids=[
                question_id,
            ],
            # magic string to return more fields
            #       !9_bDDx5Ia
            # answers
            # comments
            # body
            # body_markdown
            # tags
            filter="!--1nZwHGSSZl",
        ),
    )


@lru_cache(maxsize=1000)
def get_json_by_answer_id(answer_id: int) -> Dict[str, Any]:
    """Low level access, returns unprocessed json"""
    return cast(
        Dict[str, Any],
        SITE.fetch(
            "answers/{ids}",
            ids=[
                answer_id,
            ],
            # magic string to include certain fields.
            # need:
            #   link
            #   body_markdown
            #   body
            #   comments
            #   tags (of q)
            #   title (of q)
            filter="!3zl2.DbpKHRASLD)i",
        ),
    )


@lru_cache(maxsize=1000)
def get_json_by_user_id(user_id: int) -> Dict[str, Any]:
    """Low level access, returns unprocessed json"""
    return cast(
        Dict[str, Any],
        SITE.fetch(
            "users/{ids}",
            ids=[
                user_id,
            ],
            # just stuff we care about + about_me
            filter="!)si8a_4RZpJGdK21mxCq",
        ),
    )


# /2.2/posts/26344315/revisions?site=stackoverflow
@lru_cache(maxsize=1000)
def get_json_revisions_by_post_id(post_id: int) -> Dict[str, Any]:
    """
    Low level access, returns unprocessed json
    A post id is a question OR an answer id!
    """
    return cast(
        Dict[str, Any],
        SITE.fetch(
            "posts/{ids}/revisions",
            ids=[
                post_id,
            ],
            filter="!9_bDDm2Bc",
        ),
    )


@lru_cache(maxsize=1000)
def get_json_comments_by_post_id(post_id: int) -> Dict[str, Any]:
    """Low level access, returns unprocessed json"""
    return cast(
        Dict[str, Any],
        SITE.fetch(
            "posts/{ids}/comments",
            ids=[
                post_id,
            ],
            # need comment body & body_md, link
            filter="!--1nZxT00Un.",
        ),
    )

@lru_cache(maxsize=1000)
def get_json_related_tags(tag: str) -> Dict[str, Any]:
    """Low level access, returns unprocessed json"""
    return cast(
        Dict[str, Any],
        SITE.fetch(
            "tags/{ids}/related",
            ids=[
                tag,
            ],
        ),
    )



"""
What I really want to know is if a package exists and pypi's xmlrpc api's
are down today & at risk of going away altogether. The HTML API is ugly.

So I'm using a stats API. As a side effect, I can add a cut off for
unpopular packages.
"""

import json
from functools import lru_cache
from typing import Optional

import pypistats

# endpoints
# -----------
# recent downloads
# overall downloads, broken down by mirrors
# major, minor, system, downloads by minor version, system, etc
import requests


@lru_cache(maxsize=1000)
def get_download_count(module: str) -> Optional[int]:
    """Get download count and cache it"""
    try:
        item_string = pypistats.overall(module.strip(), format="json")
    except requests.exceptions.HTTPError as error:
        if error.response.status_code == 404:
            return None
        if error.response.status_code == 429:
            # too many requests
            # BUG: This is wrong
            return None
        raise
    # print(item_string)
    item = json.loads(item_string)
    downloads = None
    if isinstance(item, dict):
        for category in item["data"]:
            if "without_mirrors" in category.values():
                downloads = category["downloads"]
    return downloads



# noinspection PyPep8
"""so_pip/StackOverflow Pip
Not associated with PyPA, nor StackOverflow.

Usage:
  so_pip vendorize <name> (--question=<question_id>|--answer=<answer_id>|--package=<package>) [--revision=<revision>] [options]
  so_pip search <name> --query=<query> --tags=<tags> [--count=<count>] [options]
  so_pip uninstall <names>... [options]
  so_pip list [options]
  so_pip freeze [options]
  so_pip show <names>... [options]
  so_pip (-h | --help)
  so_pip --version

Options:
  -h --help                    Show this screen.
  -v --version                 Show version.
  -c --count=<count>           How many posts to get [default: 2].
  -o --output=<output>         Folder for packages. Defaults to /output
  -q --question=<question_id>  Stackoverflow question id
  -a --answer=<answer_id>      Stackoverflow answer id
  -r --revision=<revision>     Revision id for answer.
  --verbose                    Show logging
  --quiet                      No informational logging

"""
import logging
import sys

import docopt

from so_pip import _version as meta
from so_pip import settings as settings
from so_pip.commands import freeze as freeze
from so_pip.commands import list_all as list_all
from so_pip.commands import search as search
from so_pip.commands import show as show
from so_pip.commands import uninstall as uninstall
from so_pip.commands import vendorize as vendorize

# Do these need to stick around?
LOGGERS = []

LOGGER = logging.getLogger(__name__)


def main() -> int:
    """Get the args object from command parameters"""
    arguments = docopt.docopt(__doc__, version=f"so_pip {meta.__version__}")
    # print(arguments)
    # LOGGER.debug(arguments)
    output_folder = arguments["--output"]
    if arguments["--quiet"]:
        settings.QUIET = True

    if arguments["--verbose"]:
        # root logger, all modules
        for root in ("so_pip", "__main__", "url_lib3"):
            logger = logging.getLogger(root)
            logger.setLevel(logging.DEBUG)
            handler = logging.StreamHandler()
            handler.setLevel(logging.DEBUG)
            log_format = "%(asctime)s - %(name)s - %(levelname)s - %(message)s"
            formatter = logging.Formatter(log_format)
            handler.setFormatter(formatter)
            logger.addHandler(handler)
            LOGGERS.append(logger)
    if arguments["vendorize"]:
        prefix = arguments["<name>"] or ""
        question = arguments["--question"]
        answer = arguments["--answer"]
        revision = arguments["<revision>"]
        if not question and not answer:
            print("Must specify --question or --answer identifier")
            return -1
        if not output_folder:
            print("No --output folder specified")
            return -1
        if question:
            packages_made = vendorize.import_so_question(
                prefix, question, output_folder
            )
        elif answer:
            packages_made = vendorize.import_so_answer(
                prefix, answer, output_folder, revision
            )
        else:
            raise TypeError("Need to specify a question or answer")
        print(f"Vendorized {','.join(packages_made)} at {output_folder}")
    elif arguments["uninstall"]:
        packages = arguments["<name>"]
        if not packages:
            print("No packages specified to uninstall")
            return -1
        for package in packages:
            uninstall.uninstall_package(output_folder, package)
        print(
            f"Uninstalled {','.join(packages)} from vendorized folder.\n"
            f"If you also installed with pip you will need to uninstall with pip"
        )
    elif arguments["list"]:
        if not output_folder:
            print("Missing --output folder")
            return -1
        list_all.list_packages(output_folder)
    elif arguments["freeze"]:
        if not output_folder:
            print("Missing --output folder")
            return -1
        freeze.freeze_environment(output_folder)
    elif arguments["show"]:
        packages = arguments["<names>"]
        if not packages:
            print("No packages specified to show")
            return -1
        for package in packages:
            show.show(output_folder, package)
    elif arguments["search"]:
        prefix = arguments["<name>"]
        if not prefix:
            prefix = ""
        query = arguments["--query"]
        if not query:
            print("--query required for search")
            return -1

        try:
            count_str = arguments["--count"]
            count = int(count_str)
        except ValueError:
            print(f"Can't convert {arguments.get('--count', '')} to a number")
            return -1

        # TODO: better way to do this with docopts
        if arguments["--tags"]:
            tags = arguments["--tags"].split(";,")
        else:
            tags = []
        search.import_so_search(prefix, query, tags, output_folder, count)
    else:
        print("Don't recognize that command.")
        return -1
    return 0


if __name__ == "__main__":
    sys.exit(main())



"""
Shell out to run a few commands

- pytest
- pip-upgrade
- pur
- safety
- pyflakes
- pipreqs
- futurize
- 2to3
- black
- pyupgrade
- isort
- pylint
- pypinfo
- vermin
"""
import logging
import os
import shlex

from so_pip import settings as settings
from so_pip.cli_clients.subprocess_utils import execute_get_text
from so_pip.utils import guards as guards
from so_pip.utils.files_utils import find_file

LOGGER = logging.getLogger(__name__)


def pytest_detect(file: str) -> str:
    """How many tests in this file"""
    guards.must_be_truthy(file, "file required")
    # so few SO questions have executable unit tests

    command_text = f"pytest {file} --collect-only".strip().replace("  ", " ")
    LOGGER.debug(command_text)
    command = shlex.split(command_text)
    result = execute_get_text(command)
    return result


# https://stackoverflow.com/questions/24764549/
# upgrade-python-packages-from-requirements-txt-using-pip-command
def pip_upgrade(file: str) -> str:
    """
    Get latest versions & pip
    """
    guards.must_be_truthy(file, "file required")
    # pip-upgrade
    command_text = f"pip-upgrade --default-index-url --file {file}".strip().replace(
        "  ", " "
    )
    LOGGER.debug(command_text)
    command = shlex.split(command_text)
    result = execute_get_text(command)
    return result


def pur(file: str) -> str:
    """
    Get latest versions & pip
    """
    guards.must_be_truthy(file, "file required")
    # alternative... this one upgrades pinned, too
    # https://github.com/alanhamlett/pip-update-requirements

    command_text = f"pur -f --requirement {file}".strip().replace("  ", " ")
    LOGGER.debug(command_text)
    command = shlex.split(command_text)
    result = execute_get_text(command)
    return result


def safety(file: str) -> str:
    """Check if dep is malicious/insecure"""
    guards.must_be_truthy(file, "file required")
    command = shlex.split(f"safety check --file {file}".strip().replace("  ", " "))
    LOGGER.debug(command)
    result = execute_get_text(command)
    return result


def pyflakes(file: str) -> str:
    """Just run pyflakes"""
    guards.must_be_truthy(file, "file required")
    command = shlex.split(
        f"{settings.SHELL} pyflakes {file}".strip().replace("  ", " ")
    )
    LOGGER.debug(command)
    result = execute_get_text(command)
    return result


def generate_requirements(folder: str) -> str:
    """Make more installable"""
    guards.must_be_truthy(folder, "folder required")
    command = shlex.split(
        f"{settings.SHELL} pipreqs {folder} --force".strip().replace("  ", " ")
    )
    LOGGER.debug(command)
    result = execute_get_text(command)
    return result


def futurize(file_name: str) -> str:
    """Yet another py2 to 3 converter"""
    guards.must_be_truthy(file_name, "file_name required")
    text = f"{settings.SHELL} futurize --stage1 -w {file_name}"
    LOGGER.debug(text)
    command = shlex.split(text)
    result = execute_get_text(command)
    LOGGER.debug(result)
    return result


def black(folder_name: str) -> str:
    """Format files to keep pylint happy"""
    guards.must_be_truthy(folder_name, "folder_name required")
    text = f"{settings.SHELL} black {folder_name} --target-version=py38"
    LOGGER.debug(text)
    command = shlex.split(text)
    result = execute_get_text(command, ignore_error=True)
    LOGGER.debug(result)
    return result


def two_to_three(file_name: str) -> str:
    """fix print"""
    guards.must_be_truthy(file_name, "file_name required")
    text = f"{settings.SHELL} 2to3 -w {file_name}"
    LOGGER.debug(text)
    command = shlex.split(text)
    result = execute_get_text(command)
    LOGGER.debug(result)
    return result


def pyupgrade(file_name: str) -> str:
    """Bump to 3.7+"""
    guards.must_be_truthy(file_name, "file_name required")
    command = (
        f"{settings.SHELL} pyupgrade "
        f"--py37-plus "
        f"--exit-zero-even-if-changed {file_name}".strip().replace("  ", " ")
    )
    LOGGER.debug(command)
    parts = shlex.split(command)

    result = execute_get_text(parts, ignore_error=False)
    LOGGER.debug(result)
    return result


def isort(folder: str) -> str:
    """Sort the imports to discover import order bugs and prevent import order bugs"""

    guards.must_be_truthy(folder, "folder required")
    # This must run before black. black doesn't change import order but it wins
    # any arguments about formatting.
    # isort MUST be installed with pipx! It is not compatible with pylint in the same
    # venv. Maybe someday, but it just isn't worth the effort.

    command = f"{settings.SHELL} isort --profile black {folder}"
    LOGGER.debug(command)
    parts = shlex.split(command)
    try:
        result = execute_get_text(parts)
        LOGGER.debug(result)
        return result
    except FileNotFoundError:
        LOGGER.debug("isort not installed, skipping isort")
        return ""
    return ""


def pylint(folder: str) -> str:
    """Sort the imports to discover import order bugs and prevent import order bugs"""
    guards.must_be_truthy(folder, "folder required")
    # This must run before black. black doesn't change import order but it wins
    # any arguments about formatting.
    # isort MUST be installed with pipx! It is not compatible with pylint in the same
    # venv. Maybe someday, but it just isn't worth the effort.
    rcfile = find_file("pylintrc.ini", __file__)

    command = (
        f"{settings.SHELL} pylint "
        "--msg-template='{path}:{line}: [{msg_id}({symbol}), {obj}] {msg}' "
        f"--exit-zero --rcfile='{rcfile}' {folder}".strip().replace("  ", " ")
    )
    "".strip().replace("  ", " ")
    LOGGER.debug(command)
    parts = shlex.split(command)
    try:
        result = execute_get_text(parts)
        LOGGER.debug(result)
        return result
    except FileNotFoundError:
        LOGGER.debug("isort not installed, skipping isort")
        return ""
    return ""


def pypinfo(package: str) -> str:
    """
    Pypi info
    """
    guards.must_be_truthy(package, "package required")
    # https://pypi.org/project/pypinfo/
    if not os.environ.get("GOOGLE_APPLICATION_CREDENTIALS", None):
        return (
            "You must set up credentials "
            "to run pypinfo https://pypi.org/project/pypinfo/"
        )
    command = f"{settings.SHELL} pypinfo {package}"
    LOGGER.debug(command)
    parts = shlex.split(command)
    result = execute_get_text(parts, ignore_error=True)
    LOGGER.debug(result)
    return result


def vermin(package: str) -> str:
    """
    vermin info
    """
    guards.must_be_truthy(package, "package required")

    command = f"{settings.SHELL} vermin {package}"
    LOGGER.debug(command)
    parts = shlex.split(command)
    result = execute_get_text(parts, ignore_error=True)
    LOGGER.debug(result)
    return result



"""
Uninstall command.

Tricky if using question id & post id
"""
import os
import shutil

from so_pip.utils import guards as guards
from so_pip.utils.user_trace import inform


def uninstall_package(target_folder: str, package_name: str) -> None:
    """Clean out that old folder"""
    guards.must_be_truthy(target_folder, "target_folder required")
    guards.must_be_truthy(package_name, "package_name required")
    inform(f"Using {target_folder} as location for vendorized packages.")
    path = os.path.join(target_folder, package_name)
    if not os.path.exists(path):
        print(f"WARNING: skipping {package_name} as it is not installed.")
        return
    shutil.rmtree(path)
    return



"""
Just find file. Needs to be independent to avoid circular imports
"""
import os


def find_file(file_name: str, executing_file: str) -> str:
    """
    Create/find a valid file name relative to a source file, e.g.
    find_file("foo/bar.txt", __file__)
    """
    file_path = os.path.join(
        os.path.dirname(os.path.abspath(executing_file)), file_name
    ).replace("\\", "/")
    return file_path



"""
Separate model for authors, licenses, maybe changelog
"""
import logging
import re
from dataclasses import dataclass, field
from typing import Any, Dict, List, Optional

from so_pip.api_clients.stackapi_facade import (
    get_json_by_user_id,
    get_json_comments_by_post_id,
    get_json_revisions_by_post_id,
)
from so_pip.api_clients.stackoverflow_scraper import scrape_urls

LOGGER = logging.getLogger(__name__)


@dataclass()
class Contribution:
    """Represents what a user did"""

    contribution_type: str = ""
    contribution_date: str = ""
    # ignore dual licenses
    contribution_license: str = ""


@dataclass()
class Author:
    """Represents someone who wrote a Q, A, comment or revision"""

    author_id: int = 0
    # only if in bio!
    emails: List[str] = field(default_factory=list)
    # twitter, github, SO, homepage
    urls: List[str] = field(default_factory=list)
    homepage: str = ""
    twitter: str = ""
    github: str = ""
    display_name: str = ""
    # e.g. original question, answer, question edit, answer edit, comment
    roles: List[str] = field(default_factory=list)
    contributions: List[Contribution] = field(default_factory=list)


@dataclass()
class Authors:
    """List of Authors"""

    question_id: int = 0
    answer_id: int = 0
    everyone: List[Author] = field(default_factory=list)


def email_from_bio(bio: str) -> List[str]:
    """Rationale-- if it is not obfuscated, then they don't mind it being public"""
    # https://meta.stackexchange.com/users/98786/robert-cartaino
    matches = re.findall(r"([a-zA-Z0-9._-]+@[a-zA-Z0-9._-]+\.[a-zA-Z0-9_-]+)", bio)
    return matches


def normalize_user_link(url: str, user_id: int) -> str:
    """Strip off end e.g. 123/user_name"""
    return f"{url.split(str(user_id))[0]}{user_id}"


def bind_question_to_authors(question: Dict[str, Any]) -> Authors:
    """get authors for question"""
    authors = Authors()
    authors.question_id = question["question_id"]
    add_authors_from_post(question, authors, is_answer=False)
    return authors


def bind_answer_to_authors(
    answer: Dict[str, Any], question: Optional[Dict[str, Any]]
) -> Authors:
    """Get authors for post"""
    authors = Authors()
    authors.answer_id = answer["answer_id"]

    add_authors_from_post(answer, authors, is_answer=True)
    if question:
        authors.question_id = answer["question_id"]
        add_authors_from_post(answer, authors, is_answer=False)
    return authors


def add_authors_from_post(post: Dict[str, Any], authors: Authors, is_answer: bool):
    """Add authors from post to Object"""
    if post["owner"]:
        owner = Author()
        if is_answer:
            owner.roles.append("Answer Owner")
        else:
            owner.roles.append("Question Owner")
        if "user_id" in post["owner"]:
            owner.twitter, owner.github = scrape_urls(post["owner"]["user_id"])
            if owner.twitter:
                owner.urls.append(owner.twitter)
            if owner.github:
                owner.urls.append(owner.github)

            full_user = get_json_by_user_id(post["owner"]["user_id"])["items"][0]
            if "website_url" in full_user and full_user["website_url"]:
                owner.urls.append(full_user["website_url"])
            if "about_me" in full_user and full_user["about_me"]:
                emails = email_from_bio(full_user["about_me"])
                if emails:
                    owner.emails.extend(emails)
            if "display_name" in post["owner"] and post["owner"]["display_name"]:
                owner.display_name = post["owner"]["display_name"]
            else:
                owner.display_name = "Name not available"
        else:
            owner.display_name = "Name not available"

        if "link" in post["owner"] and post["owner"]["link"]:
            url = normalize_user_link(post["owner"]["link"], post["owner"]["user_id"])
            if url:
                owner.urls.append(url)
        authors.everyone.append(owner)
    else:
        if post["user"]["user_type"] != "does_not_exist":
            LOGGER.debug("What sort of user is this?")

    post_id_name = "answer_id" if is_answer else "question_id"
    revision_json = get_json_revisions_by_post_id(post[post_id_name])
    for revision in revision_json.get("items", []):
        if revision["user"]["user_type"] == "does_not_exist":
            continue
        reviser = Author()
        if is_answer:
            reviser.roles.append("Answer Reviser")
        else:
            reviser.roles.append("Question Reviser")
        if "display_name" in revision["user"] and revision["user"]["display_name"]:
            reviser.display_name = revision["user"]["display_name"]
        else:
            reviser.display_name = "Name not available"
        full_user = get_json_by_user_id(revision["user"]["user_id"])["items"][0]
        if "website_url" in full_user and full_user["website_url"]:
            reviser.urls.append(full_user["website_url"])
        if "about_me" in full_user and full_user["about_me"]:
            email=email_from_bio(full_user["about_me"])
            if email:
                reviser.emails.extend(email)
        if "link" in post["owner"] and post["owner"]["link"]:
            url = normalize_user_link(
                revision["user"]["link"], revision["user"]["user_id"]
            )
            if url:
                reviser.urls.append(url)

        authors.everyone.append(reviser)
    comments = get_json_comments_by_post_id(post[post_id_name])
    for comment in comments.get("items", []):
        if comment["owner"]["user_type"] == "does_not_exist":
            continue
        commenter = Author()
        if is_answer:
            commenter.roles.append("Answer Commenter")
        else:
            commenter.roles.append("Question Commenter")
        if "display_name" in comment["owner"] and comment["owner"]["display_name"]:
            commenter.display_name = comment["owner"]["display_name"]
        else:
            commenter.display_name = "Name not available"
        full_user = get_json_by_user_id(comment["owner"]["user_id"])["items"][0]
        if "website_url" in full_user and full_user["website_url"]:
            commenter.urls.append(full_user["website_url"])
        if "about_me" in full_user and full_user["about_me"]:
            email = email_from_bio(full_user["about_me"])
            if email:
                commenter.emails.extend(email)
        if "link" in comment["owner"] and comment["owner"]["link"]:
            url = comment["owner"]["link"]
            link = normalize_user_link(url, comment["owner"]["user_id"])
            if link:
                commenter.urls.append(link)
        authors.everyone.append(commenter)



"""
Create pyproject.toml file

Supports:
- python version from either minver or so_pip.ini
- list of inferred dependencies
- list of tags
"""
import logging
from typing import Any, Dict, Optional

import toml

from so_pip.make_from_template import load_template
from so_pip.models.python_package_model import CodePackage

LOGGER = logging.getLogger(__name__)


def create_pytroject_toml(
    package_folder: str,
    python_submodule: CodePackage,
    question: Dict[str, Any],
    answer: Optional[Dict[str, Any]],
) -> None:
    """Put everything into setup.cfg"""
    # maybe already done?
    # post_id = answer["answer_id"] if answer else question["question_id"]
    # revisions_json = get_json_revisions_by_post_id(post_id).get("items", [])
    if answer:
        url = answer["link"]
    else:
        url = question["link"]

    data = {
        "dependencies": python_submodule.dependencies,
        "question_title": question["title"],
        "revision_number": python_submodule.version,  # or len(revisions_json)
        "minimum_python": python_submodule.minimum_python,
        "name": python_submodule.package_name,
        "tags": question["tags"],
        "url": url,
    }

    output_text = render_pyproject_toml(data)

    if not output_text:
        raise TypeError("Expected output_text by now")
    with open(
        package_folder + "/pyproject.toml", "w", encoding="utf-8", errors="replace"
    ) as project_file:
        project_file.write(output_text)
        # see if it is valid toml
        # pylint: disable=broad-except
        # noinspection PyBroadException
        try:
            toml.loads(output_text)
        except BaseException as exception:
            LOGGER.error("parse failed " + str(exception))
            raise


def render_pyproject_toml(
    data: Dict[str, Any],
) -> str:
    """
    Help the world get away from setup.py
    """
    template = load_template(template_filename="pyproject.toml.jinja", autoescape=False)
    # Turn off autoescape because this is python not html.
    output_text = template.render(data=data, autoescape=False)  # nosec
    # validate toml
    _ = toml.loads(output_text)
    return output_text



"""
List packages
"""
import os
from typing import List

from so_pip.utils import guards as guards
from so_pip.utils.user_trace import inform


def list_dirs(path: str) -> List[str]:
    """Get list of directories"""
    # https://stackoverflow.com/a/31049707/33264
    guards.must_be_truthy(path, "path required")
    return [
        directory
        for directory in os.listdir(path)
        if os.path.isdir(os.path.join(path, directory))
    ]


def list_packages(target_folder: str, quiet: bool = False) -> None:
    """List packages"""
    guards.must_be_truthy(target_folder, "target_folder required")
    if not quiet:
        inform(f"Using {target_folder} as vendorized folder")
    path = os.path.join(target_folder)
    for folder in list_dirs(path):
        # TODO: Add version
        print(folder)



""""
Add safe comments symbols to any list of string
"""
import html2text
from bs4 import BeautifulSoup
from so_pip.parse_python.code_transformations import html_to_python_comments

LANGUAGE_TO_SYMBOL = {
    "batchfile": [".bat", "REM"],
    "c": [".c", "//"],
    "c#": [".cs", "//"],
    "c++": [".cpp", "//"],
    "css": [".css", "//"],
    "coffeescript": [".coffee", "//"],
    "erlang": [".erlang", "//"],
    "go": [".go", "//"],
    "html": [".html", ""],
    "haskell": [".haskell", "//"],
    "java": [".java", "//"],
    "javascript": [".js", "//"],
    "jupyter notebook": [".nb", ""],
    "lua": [".lua", "//"],
    "markdown": [".md", ""],
    "matlab": [".matlab", "//"],
    "objective-C": [".objc", "//"],
    "php": [".php", "//"],
    "perl": [".perl", "//"],
    "powershell": [".ps1", "#"],
    "python": [".py", "#"],
    "r": [".r", "#"],
    "ruby": [".rb", "#"],
    "rust": [".rust", "//"],
    "sql": [".sql", "--"],
    "scala": [".scala", "//"],
    "shell": [".sh", "#"],
    "swift": [".swift", "//"],
    "tex": [".tex", "//"],
    "typescript": [".ts", "//"],
}

def html_to_comments(html, language:str)->str:
    """Convert html to text and comment it out according to
    line by line comment syntax for language"""
    if language in ("python", ".py"):
        return html_to_python_comments(html)

    soup = BeautifulSoup(html, features="html.parser")
    check_if_blank = soup.get_text()
    if check_if_blank.isspace():
        # black will add space if we need it
        return ""

    # textwrap.wrap makes a mess
    # black does not re-format overlong comment lines.
    # textwrap.wrap(text, replace_whitespace=False)
    text = html2text.html2text(str(soup))
    text = text.strip(" \n")
    return comment_lines(text, language)

def comment_lines(code: str, language: str) -> str:
    """Add comments for any language in a naive way."""
    lines = code.split("\n")
    symbol = "//"
    if language != "." and "." in language:
        extension = language.lower().split(".")[1]
        # TODO: this a waste of cpu but still probably quick
        for _, value in LANGUAGE_TO_SYMBOL.items():
            if extension == value[0]:
                symbol = value[1]
    else:
        symbol = LANGUAGE_TO_SYMBOL.get(language, ["", ""])[1]

    lines = [symbol + " " + line for line in lines]
    return "\n".join(lines)



"""
Reduce friction of working with subprocess
"""
import logging
import subprocess  # nosec
from typing import Dict, List, Optional

LOGGER = logging.getLogger(__name__)


def execute_get_text(
    command: List[str],
    ignore_error: bool = False,
    # shell: bool = True, # causes cross plat problems, security warnings, etc.
    env: Optional[Dict[str, str]] = None,
) -> str:
    """
    Execute shell command and return stdout txt
    """

    completed = None
    try:
        completed = subprocess.run(  # nosec
            command,
            check=not ignore_error,
            # shell=shell, # causes cross plat problems, security warnings, etc.
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            env=env,
        )
    except subprocess.CalledProcessError as cpe:
        if ignore_error and completed:
            return completed.stdout.decode("utf-8") + completed.stderr.decode("utf-8")
        LOGGER.debug(cpe)
        raise
    else:
        return completed.stdout.decode("utf-8") + completed.stderr.decode("utf-8")



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['so_pip',
 'so_pip._cache',
 'so_pip._vendor',
 'so_pip._vendor.find_imports',
 'so_pip._vendor.find_imports_cruft',
 'so_pip.api_clients',
 'so_pip.cli_clients',
 'so_pip.commands',
 'so_pip.models',
 'so_pip.parse_code',
 'so_pip.parse_python',
 'so_pip.pypi_query',
 'so_pip.random_names',
 'so_pip.random_names.etl',
 'so_pip.support_files',
 'so_pip.update_any',
 'so_pip.utils']

package_data = \
{'': ['*'],
 'so_pip': ['licenses/*', 'templates/*'],
 'so_pip._vendor': ['find_imports_cruft/LICENSE/*']}

install_requires = \
['2to3',
 'beautifulsoup4',
 'black',
 'html2text',
 'jinja2',
 'markdown',
 'nbformat',
 'pip-upgrader',
 'pipreqs',
 'pur',
 'py-stackexchange',
 'pyflakes',
 'pypinfo',
 'pypistats',
 'pyrankvote',
 'python-dotenv',
 'pyupgrade',
 'requests',
 'stackapi',
 'stdlib-list',
 'toml',
 'vermin',
 'whats-that-code>=0.1.4']

entry_points = \
{'console_scripts': ['so_pip = so_pip.__main__:main']}

setup_kwargs = {
    'name': 'so-pip',
    'version': '0.1.29',
    'description': 'Generate packages from Stackoverflow answers',
    'long_description': 'so_pip\n======\nEveryone copies code from StackOverflow, but no one is formalizing it.\n\nThis will vendorize the source code of question or answer into a folder and\ngenerate the accessory files to make it look like a python package.\n\nThe feature-set overlaps a bit with [cookie cutter, vendorizing libraries and\nstackoverflow search cli\'s](https://github.com/matthewdeanmartin/so_pip/blob/main/docs/prior_art.md).\n\nInstallation\n------------\nRequires Python 3.7+, tested with 3.7, 3.8, 3.9\n```\npip install so_pip\nso_pip vendorize my_name --question=31049648 --output=output\n```\n\nUsing via [dockerhub](https://hub.docker.com/repository/docker/matthewdeanmartin/so_pip)\n```\n# for mac, unix, cmd.exe, powershell\ndocker pull matthewdeanmartin/so_pip\ndocker run --rm -i -v "$PWD/data:/data" matthewdeanmartin/so_pip --help\n```\nIf you use git bash/mingw64/cygwin, see [run.sh](https://github.com/matthewdeanmartin/so_pip/blob/main/docker/run.sh)\nbecause docker needs help doing a volume mount.\n\n\nUsage\n--------------\nConsider getting a [key](https://stackapps.com/apps/oauth/register) and adding a [.so_pip.ini file](https://github.com/matthewdeanmartin/so_pip/blob/main/.so_pip.ini) The app will make best efforts if you don\'t.\n```\n# Turn posts into nicely formated packages\n> so_pip vendorize my_name --question=31049648 | --answer=31049648\n> so_pip search --answer=31049648 --tags=python\n\n# Pip-like commands\n> so_pip uninstall | upgrade {package_name}\n> so_pip list | freeze\n```\n\nDocs\n-----\n* [Examples](https://github.com/matthewdeanmartin/so_pip/tree/main/examples)\n* [CLI](https://github.com/matthewdeanmartin/so_pip/blob/main/docs/cli.md)\n* [Code reuse scanarios you see on StackOverflow](https://github.com/matthewdeanmartin/so_pip/blob/main/docs/scenarios.md)\n* [Features](https://github.com/matthewdeanmartin/so_pip/blob/main/docs/features.md)\n* [Security Considerations](https://github.com/matthewdeanmartin/so_pip/blob/main/docs/security.md)\n* [Prior Art](https://github.com/matthewdeanmartin/so_pip/blob/main/docs/prior_art.md) Similar and overlapping tools.\n* [Contributing *answers* to StackOverflow](https://github.com/matthewdeanmartin/so_pip/blob/main/docs/contributing.md) AKA, fixing answers you found.\n* [Attribution Compliance](https://github.com/matthewdeanmartin/so_pip/blob/main/docs/comply_with_cc_sa.md)\n* [Contributing to so_pip](https://github.com/matthewdeanmartin/so_pip/blob/main/CONTRIBUTING.md)\n* [Code of Conduct for so_pip](https://github.com/matthewdeanmartin/so_pip/blob/main/CODE_OF_CONDUCT.md)\n',
    'author': 'Matthew Martin',
    'author_email': 'matthewdeanmartin@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/matthewdeanmartin/so_pip',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.6',
}


setup(**setup_kwargs)



"""
Create a code of conduct file
"""
import markdown

from so_pip.make_from_template import load_template


def render_code_of_conduct(package_folder: str) -> str:
    """
    Links to SO's code of conduct
    """
    template = load_template(
        template_filename="CODE_OF_CONDUCT.md.jinja", autoescape=False
    )
    output_text = template.render()
    # check if markdown is okay
    _ = markdown.markdown(output_text)
    with open(
        package_folder + "/CODE_OF_CONDUCT.md", "w", encoding="utf-8", errors="replace"
    ) as author_file:
        author_file.write(output_text)

    return output_text



"""
Authors file

This is roughly following Gnu Gnit's conventions.

Also draws inspiration from google's format.
https://opensource.google/docs/releasing/authors/

Also draws inspiration from libraries that attempt to parse AUTHORS files.
https://www.npmjs.com/package/stringify-author
https://www.npmjs.com/package/author-regex
"""
import logging
from collections import defaultdict
from typing import Any, Collection, DefaultDict, Dict, List, Optional, Sequence, Union

import markdown

from so_pip.make_from_template import load_template
from so_pip.models.authors_model import (
    Author,
    bind_answer_to_authors,
    bind_question_to_authors,
)

LOGGER = logging.getLogger(__name__)


def write_authors(
    package_folder: str,
    package_name: str,
    question: Dict[str, Any],
    answer: Optional[Dict[str, Any]] = None,
) -> None:
    """write file"""
    if answer:
        authors = bind_answer_to_authors(answer, question)
    else:
        authors = bind_question_to_authors(question)

    sections: DefaultDict[str, List[Author]] = defaultdict(list)
    for author in authors.everyone:
        for role in author.roles:
            if author not in sections[role]:
                sections[role].append(author)
    for _, value in sections.items():
        value.sort(key=lambda _: _.display_name.lower())
    for author in authors.everyone:
        author.urls.sort(key=lambda _: "" if "stackoverflow.com" in _ else _.lower())

    with open(
        package_folder + "/AUTHORS.md", "w", encoding="utf-8", errors="replace"
    ) as author_file:
        item = {"name": package_name, "authors": sections}
        text = render_authors(data=item)
        # calling this to see if it is somewhat valid markdown
        _ = markdown.markdown(text)
        author_file.write(text)


def render_authors(
    data: Dict[str, Union[str, Sequence[str], Collection[str]]],
) -> str:
    """
    Render minimal setup.py suitable for `pip install -e .`
    """
    template = load_template(template_filename="AUTHORS.jinja", autoescape=False)
    output_text = template.render(item=data)
    return output_text



"""
Install one question or post.

Vendorize because I'm not installing it to a venv.
"""

import collections
from typing import Any, Dict, List, Optional, Tuple

from so_pip import settings as settings
from so_pip.api_clients import stackapi_facade as stackapi_client
from so_pip.cli_clients.external_commands import black, isort, pur, pylint, safety
from so_pip.file_writing import write_as_html, write_as_md, write_as_text
from so_pip.models.code_file_model import CodeFile
from so_pip.models.python_package_model import CodePackage
from so_pip.parse_code.write_anything import write_and_format_any_file
from so_pip.parse_python.format_code import write_and_format_python_file
from so_pip.parse_python.make_reusable import is_reusable
from so_pip.parse_python.module_maker import (
    create_package_folder,
    map_post_to_code_package_model,
)
from so_pip.parse_python.python_validator import validate_with_vermin
from so_pip.parse_python.upgrade_to_py3 import upgrade_file
from so_pip.random_names.make_name import make_up_module_name
from so_pip.settings import KEEP_ANSWERS_WITH_NO_CODE, KEEP_ANSWERS_WITH_NO_DEF_OR_CLASS
from so_pip.support_files.authors import write_authors
from so_pip.support_files.changelog import changelog_for_post
from so_pip.support_files.code_of_conduct import render_code_of_conduct
from so_pip.support_files.license import write_license
from so_pip.support_files.pyproject_toml import create_pytroject_toml
from so_pip.support_files.readme_md import create_readme_md
from so_pip.support_files.requirements_for_post import requirements_for_file
from so_pip.utils.user_trace import inform


def handle_post(
    output_folder: str,
    package_prefix: str,
    question: Dict[str, Any],
    answers: List[Dict[str, Any]],
    answer_revision: Optional[Dict[str, Any]] = None,
) -> List[str]:
    """Loop through answers"""
    # HACK: still thinking this out
    if answer_revision and len(answers) == 1:
        answers[0]["body"] = answer_revision["body"]
        # BUG: pretty sure body_md is used too?!

    packages_made: List[str] = []

    posts: List[Tuple[str, Dict[str, Any]]] = []
    for data in answers:
        posts.append(("answer", data))

    # TODO: switch for if we include Q
    posts.append(("question", question))

    for post_type, shallow_post in posts:
        if post_type == "answer" and shallow_post["score"] < settings.MINIMUM_SCORE:
            inform(f"Answer lacks minimum score of {settings.MINIMUM_SCORE}...skipping")
            continue

        if post_type == "answer":
            post = stackapi_client.get_json_by_answer_id(shallow_post["answer_id"])[
                "items"
            ][0]
        else:
            post = question

        if (
            post_type == "answer"
            and not KEEP_ANSWERS_WITH_NO_DEF_OR_CLASS
            and not is_reusable(post["body"])
        ):
            # TODO: make this more strict
            inform("Answer lacks def/class, not re-usable...skipping")
            continue

        def post_has_code(answer: Dict[str, Any]) -> bool:
            """This will probably get more complicated"""
            return "<pre><code" in answer["body"] and "</code>" in answer["body"]

        if (
            not post_has_code(post)
            and not KEEP_ANSWERS_WITH_NO_CODE
            and post_type == "answer"
        ):
            inform("Answer lacks code blocks,...skipping")
            continue

        if post_type == "answer":
            module_name = make_up_module_name(post["answer_id"], package_prefix, "a")
        else:
            module_name = make_up_module_name(post["question_id"], package_prefix, "q")
        packages_made.append(module_name)

        package_info = map_post_to_code_package_model(
            post,
            post["body"],
            module_name,
            f"StackOverflow post #{question['title']}",
            tags=question["tags"],
        )

        i = 0
        package_info.extract_metadata(post)
        if settings.METADATA_IN_INIT:
            metadata_for_init = "\n".join(
                package_info.header + package_info.python_metadata
            )
        else:
            metadata_for_init = ""

        supporting_files_folder, python_source_folder = create_package_folder(
            output_folder, module_name, module_name, metadata_for_init
        )
        wrote_py_file = False

        submodule_path = f"{python_source_folder}/main"

        frequencies = package_info.file_frequencies()
        for code_file in package_info.code_files:
            i += 1
            success = write_one_code_file(
                code_file, frequencies, i, package_info, submodule_path, joiner=""
            )
            if success:
                wrote_py_file = True

        if settings.POSTS_AS_HTML:
            write_as_html(post, f"{supporting_files_folder}/post")
        if settings.POSTS_AS_MD:
            write_as_md(post, f"{supporting_files_folder}/post")
        if settings.POSTS_AS_TXT:
            write_as_text(post, f"{supporting_files_folder}/post")

        write_license(post, supporting_files_folder)
        if settings.GENERATE_CHANGE_LOG:
            changelog_for_post(post, supporting_files_folder)
        if settings.GENERATE_AUTHORS:
            write_authors(
                supporting_files_folder,
                module_name,
                question,
                post if post_type == "answer" else None,
            )
        if settings.GENERATE_README:
            create_readme_md(
                supporting_files_folder,
                package_info,
                question,
                post if post_type == "answer" else None,
            )
        if settings.GENERATE_CODE_OF_CONDUCT:
            render_code_of_conduct(supporting_files_folder)

        if wrote_py_file:
            python_versions = validate_with_vermin(python_source_folder)
            package_info.minimum_python = python_versions
            print(package_info.minimum_python)

            # extract requirements, pin, check for security issues
            requirements_txt, count = requirements_for_file(
                supporting_files_folder, package_info
            )
            create_pytroject_toml(supporting_files_folder, package_info, question, post)
            if requirements_txt and count > 0:
                pur(requirements_txt)
                result = safety(requirements_txt)
                inform(result)
            # Upgrade, but this is one of the slower things to do.
            # Also, eval(input()) makes things worse.
            if settings.BUMP_TO_PY3:
                upgrade_file(submodule_path)
            # format
            isort(output_folder)
            black(output_folder)
            # lint
            lint_file_name = supporting_files_folder + "/lint.txt"
            with open(
                lint_file_name, "w", encoding="utf-8", errors="replace"
            ) as lint_writer:
                lint_writer.write(pylint(python_source_folder))

    return packages_made


def write_one_code_file(
    code_file: CodeFile,
    frequencies: collections.Counter,
    i: int,
    package_info: CodePackage,
    submodule_path: str,
    joiner: str,
) -> bool:
    """Just code to write a python file"""
    wrote_py_file = False
    file_header = (
        package_info.brief_header if settings.METADATA_IN_INIT else package_info.header
    )
    if frequencies.get(code_file.extension, 0) > 1:
        code_file_name = f"{submodule_path}_{joiner}{i}{code_file.extension}"
    else:
        code_file_name = f"{submodule_path}{code_file.extension}"

    to_write = code_file.to_write()
    if code_file.extension == ".py":
        metadata = [] if settings.METADATA_IN_INIT else package_info.python_metadata
        code_to_write = file_header + metadata + to_write
        wrote_py_file = write_and_format_python_file(
            code_file_name,
            code_to_write,
        )

    else:
        # Write a different file.
        # Same thing, but different comment symbols and no python improvements
        headers = (
            package_info.brief_header
            if settings.METADATA_IN_INIT
            else package_info.header
        )
        code_to_write = headers + to_write
        write_and_format_any_file(
            code_file_name,
            code_to_write,
        )
    return wrote_py_file



"""
Using source code, find requirements.

Removes python 3.8 system libraries.

Assumes module names match package names. This is not true.

Does not run safety if no modules are successfully guessed.
"""
import logging
import os
import subprocess  # nosec
from typing import Optional, Tuple

from stdlib_list import stdlib_list

from so_pip import settings as settings

# noinspection PyProtectedMember
from so_pip._vendor.find_imports.main import find_imports

# ^\s*(from|import)\s+\w+
from so_pip.cli_clients.external_commands import generate_requirements
from so_pip.models.python_package_model import CodePackage
from so_pip.pypi_query.main import find_modules
from so_pip.support_files.setup_cfg import create_setup_cfg

# https://github.com/ohjeah/pip-validate

LOGGER = logging.getLogger(__name__)


def requirements_for_file(
    package_folder: str, python_submodule: CodePackage
) -> Tuple[Optional[str], int]:
    """Requirements for running `safety`"""

    package_count = 0
    file_to_write = None
    all_imports = []
    with os.scandir(package_folder) as root_dir:
        for path in root_dir:
            if path.is_file():
                filename = path.name
                # for filename in os.listdir(package_folder):
                if "." not in filename:
                    raise TypeError("All files must have an extension")
                if filename.endswith(".py"):
                    py_file = path.path
                    all_imports += find_imports(py_file)
                else:
                    continue

    # remove built ins
    libraries = stdlib_list("3.8")
    python_submodule.dependencies = set(all_imports) - set(libraries)

    packages_of_same_name, not_in_pypi = find_modules(
        list(python_submodule.dependencies)
    )

    # https://stackoverflow.com/questions/8370206/how-to-get-a-list-of-built-in-modules-in-python
    if python_submodule.dependencies:
        file_to_write = package_folder + "/requirements.txt"
        with open(
            file_to_write, "w", encoding="utf-8", errors="replace"
        ) as requirements:
            requirements.write("# module names often don't match package names!\n\n")
            for _import in packages_of_same_name:
                requirements.write(_import + "\n")
                package_count += 1
            for bad_import in not_in_pypi:
                item = f"# {bad_import} # module imported but no package of same name\n"
                requirements.write(item)

    if settings.GENERATE_SETUP_CFG:
        create_setup_cfg(package_folder, python_submodule)
    return file_to_write, package_count


def process_requirements_for_a_module(package_folder: str) -> None:
    """
    Call a commandline requirements generator
    """
    if settings.GENERATE_REQUIREMENTS_TXT:
        try:
            generate_requirements(package_folder)
        except subprocess.CalledProcessError as cpe:
            LOGGER.debug(f"generate requirements failed : {package_folder}", str(cpe))



"""
Wrap in run or class
"""
from typing import List


def is_reusable(code: str) -> bool:
    """Try to find def or class"""
    # tokenizer splits only on space, not punct
    #
    # tokens = code.split()
    # TODO: when possible, actually parse the AST.
    # In the meanwhile, be very loosy goosy
    if "def" in code:
        return True
    if "class" in code:
        return True
    return False


def wrap_in_run(code: str) -> str:
    """Probably a miracle if this works for everything"""
    code = code.strip("\n ")
    imports: List[str] = []
    lines: List[str] = []
    dropped_def_run = False
    # first pass

    last_was_import = False
    run_comment = '    """"Run script"""'
    for line in code.split("\n"):
        if "import" in line.split():
            last_was_import = True
            imports.append(line)
            continue
        if line.isspace() and last_was_import:
            imports.append(line)
            continue
        if not dropped_def_run:
            if len(imports + lines) > 0 and not (imports + lines)[-1].isspace():
                lines.append("")
            lines.append("def run() -> None:")
            lines.append(run_comment)
            dropped_def_run = True
            continue
        # reduce whitespace around def run
        if line.isspace() or not line:
            if len(lines) > 0 and not lines[-1] == run_comment:
                lines.append(line)
        else:
            # non whitespace
            if dropped_def_run:
                lines.append("     " + line)
                continue
            lines.append(line)

    if len(imports + lines) > 0 and not (imports + lines)[-1].isspace():
        lines.append("")
    lines.append("if __name__ == '__main__':")
    lines.append("     run()")
    return "\n".join(imports + lines)



"""
Author: Daniel G
Author Link: https://stackoverflow.com/users/207432/daniel-g
License: CC BY-SA 2.5
Date: 1270328265
Answer Url: https://stackoverflow.com/questions/2572582/return-a-list-of-imported-python-modules-used-in-a-script/2572654#2572654
"""

__title__ = "Return a list of imported Python modules used in a script?"
__version__ = "0.1.5"
__author__ = "Daniel G"
__license__ = "CC BY-SA 2.5"
__copyright__ = "Copyright 1270328265 by Daniel G"
